import pickle
import random
import numpy as np
from numba import njit, jit
from numba.experimental import jitclass

def profile(x):
    return x

# Helper functions
def slices(start, length, n):
    """Return `n` slices starting at `start` of length `length`"""
    return [slice(i, i+length) for i in range(start, length*n + start, length)]

def bin_slices(l, n):
    """Return a list of slice() objects that split l items into n bins
    The first l%n bins are length l//n+1; the remaining n-l%n bins are length l//n
    Similar to np.array_split()"""
    return slices(0, l//n+1, l%n) + slices((l//n+1)*(l%n), l//n, n-l%n)

def topk_indices(genotype, haplotypes, n_topk):
    """Return top-k haplotype indices with fewest opposite homozygous markers compared to genotype
    The array of number of opposite homozygous loci for each haplotype (`opposite_homozygous`)
    is first shuffled so that tied values are effectively randomly sampled"""

    # Mask of homozygous loci in the genotype
    homozygous_mask = (genotype == 0) | (genotype == 2)  # note: this purposefully ignores missing loci

    # Select just these homozygous loci
    g = genotype[homozygous_mask]
    h = haplotypes[:, homozygous_mask]

    # Number of opposite homozygous loci for all haplotypes
    opposite_homozygous = np.sum(g//2 != h, axis=1)

    # Shuffle indices
    indices = np.arange(len(opposite_homozygous))
    np.random.shuffle(indices)

    # Stable argsort on the shuffled values
    args = np.argsort(opposite_homozygous[indices], kind='stable')

    # Top-k indices (fewest opposite homozygous loci)
    return indices[args[:n_topk]]


class HaplotypeLibrary(object):
    """A library of haplotypes
    Each haplotype can have an identifier (any Python object, but typically a str or int)
    The identifiers are used to select haplotypes for updating or masking
    Haplotypes should be NumPy arrays of dtype np.int8

    Some functions only work on frozen libraries; some only on unfrozen ones.
    Use freeze() and unfreeze() to swap between the two states. Typically a library is
    built with append() and then frozen to enable additional functionality"""

    def __init__(self, n_loci=None):
        self._n_loci = n_loci
        self._frozen = False
        self._haplotypes = []
        self._identifiers = []  # index to identifier mapping
        self.dtype = None


    def __repr__(self):
        return repr(self._identifiers) + '\n' + repr(self._haplotypes)


    def __len__(self):
        """Number of haplotypes in the library"""
        return len(self._haplotypes)


    def __iter__(self):
        """Iterate over tuple of (id, haplotype)"""
        for i in range(len(self)):
            yield self._identifiers[i], self._haplotypes[i]


    def append(self, haplotype, identifier=None):
        """Append a single haplotype to the library.
        Note: a copy of the haplotype is taken"""
        if self._frozen:
            raise RuntimeError('Cannot append to frozen library')

        if self.dtype is None:
            self.dtype = haplotype.dtype

        if self._n_loci is None:
            self._n_loci = len(haplotype)

        self._check_haplotype(haplotype, expected_shape=(self._n_loci,))
        self._identifiers.append(identifier)
        self._haplotypes.append(haplotype.copy())


    def freeze(self):
        """Freeze the library: convert identifier and haplotype lists to NumPy arrays"""
        if self._frozen:
            raise RuntimeError('Cannot freeze an already frozen library')
        self._haplotypes = np.array(self._haplotypes)
        self._identifiers = np.array(self._identifiers)
        self._frozen = True


    def unfreeze(self):
        """Unfreeze the library: convert identifiers and haplotypes to lists"""
        if not self._frozen:
            raise RuntimeError('Cannot unfreeze an unfrozen library')
        self._haplotypes = list(self._haplotypes)
        self._identifiers = list(self._identifiers)
        self._frozen = False


    def update(self, haplotypes, identifier):
        """Update identifier's haplotypes
        'haplotypes' can be a 1d array of loci or a 2d array of shape(#haps, #loci)"""
        if not self._frozen:
            raise RuntimeError('Cannot update an unfrozen library')
        self._check_haplotype_dtype(haplotypes)
        indices = self._indices(identifier)
        # Use Numpy's broadcasting checks to handle mismatch of shape in the following:
        self._haplotypes[indices] = haplotypes


    def exclude_identifiers(self, identifiers):
        """Return a NumPy array of haplotypes excluding specified identifiers
        'identifiers' can be a single identifier or iterable of identifiers"""
        if not self._frozen:
            raise RuntimeError('Cannot exclude from an unfrozen library')
        mask = ~np.isin(self._identifiers, identifiers)
        return self._haplotypes[mask]


    def sample_only_identifiers(self, identifiers):
        """Returns HaplotypeLibrary object of haplotypes only including specified identifiers
        'identifiers' can be a single identifier or iterable of identifiers"""
        if not self._frozen:
            raise RuntimeError('Cannot exclude from an unfrozen library')
        mask = np.isin(self._identifiers, identifiers)
        return self._sampled_library(mask)


    def sample(self, n_haplotypes):
        """Return a NumPy array of randomly sampled haplotypes"""
        if not self._frozen:
            raise RuntimeError('Cannot sample from an unfrozen library')
        if n_haplotypes > len(self):
            n_haplotypes = len(self)
        sampled_indices = np.sort(np.random.choice(len(self), size=n_haplotypes, replace=False))
        return self._haplotypes[sampled_indices]


    def sample_targeted(self, n_haplotypes, genotype, n_bins, exclude_identifiers=None):
        """Sample haplotypes that 'closely match' genotype `genotype`"""
        if not self._frozen:
            raise RuntimeError('Cannot sample from an unfrozen library')
        if n_haplotypes > len(self):
            return self

        # Get top-k in a number of marker bins
        n_topk = n_haplotypes  # unnecessary variable redifinition
        indices = np.empty((n_topk, n_bins), dtype=np.int64)
        for i, s in enumerate(bin_slices(self._n_loci, n_bins)):
            indices[:, i] = topk_indices(genotype[s], self._haplotypes[:, s], n_topk)

        # Get top n_haplotypes across the bins excluding any in exclude_ifentifiers
        sampled_indices = set()
        exclude_indices = set(self._indices(exclude_identifiers))
        for idx in indices.flatten():
            if idx not in exclude_indices:
                sampled_indices.add(idx)
            if len(sampled_indices) >= n_topk:
                break
        sampled_indices = list(sampled_indices)
        # Return HaplotypeLibrary object
        return self._sampled_library(sampled_indices)


    def exclude_identifiers_and_sample(self, identifiers, n_haplotypes):
        """Return a NumPy array of (n_haplotypes) randomly sampled haplotypes
        excluding specified identifiers.
        'identifiers' can be a single identifier or an iterable of identifiers
        Note: A copy of the haplotypes are created because of fancy indexing"""
        # Exclude specified identifiers
        if not self._frozen:
            raise RuntimeError('Cannot sample or exclude from an unfrozen library')
        exclude_mask = ~np.isin(self._identifiers, identifiers)
        n_remaining_haplotypes = exclude_mask.sum()
        # Generate random sample
        if n_haplotypes > n_remaining_haplotypes:
            n_haplotypes = n_remaining_haplotypes
        sampled_indices = np.random.choice(n_remaining_haplotypes, size=n_haplotypes, replace=False)
        sampled_indices.sort()
        # Return HaplotypeLibrary object
        return self._sampled_library(sampled_indices)


    def asMatrix(self):
        """Return the NumPy array - kept for backwards compatibility"""
        if self._frozen:
            return self._haplotypes.copy()
        return np.array(self._haplotypes)


    def removeMissingValues(self):
        """Replace missing values randomly with 0 or 1 with 50 % probability
        kept for backwards compatibility"""
        for hap in self._haplotypes:
            removeMissingValues(hap)


    def get_called_haplotypes(self, threshold = 0.99):
        """Return "called" haplotypes -- these are haplotypes which only contain integer values (0,1,9).
        For haplotypes where there is uncertainty, a threshold is used to determine whether the value is called as a value or is missing. """

        if not self._frozen:
            self.freeze()
        if self.dtype is np.int8:
            return self._haplotypes

        else:
            called_haplotypes = np.full(self._haplotypes.shape, 0, dtype = np.float32)
            for i in range(called_haplotypes.shape[0]):
                called_haplotypes[i,:] = self.call_haplotypes(self._haplotypes[i,:], threshold)
            return called_haplotypes

    @staticmethod            
    @jit(nopython=True)
    def call_haplotypes(hap, threshold):
        nLoci = len(hap)
        output = np.full(nLoci, 9, dtype = np.int8)
        for i in range(nLoci):
            if hap[i] <= 1 :
                if hap[i] > threshold : output[i] = 1
                if hap[i] < 1-threshold : output[i] = 0
        return output


    def get_haplotypes(self):
        if not self._frozen:
            self.freeze()

        return self._haplotypes

    def get_identifiers(self):
        """Get haplotype identifiers"""
        return list(self._identifiers)


    def _indices(self, identifier):
        """Get row indices associated with an identifier. These can be used for fancy indexing"""
        # Return empty list if identifier == None
        if not identifier:
            return list()
        if not self._frozen:
            raise RuntimeError("Cannot get indices from an unfrozen library")
        if identifier not in self._identifiers:
            raise KeyError(f"Identifier '{identifier}' not in library")
        return  np.flatnonzero(self._identifiers == identifier).tolist()

    def _check_haplotype_dtype(self, haplotype):
        """Check the haplotype has expected dtype"""
        if haplotype.dtype != self.dtype:
            raise TypeError('haplotype(s) not equal to library dtype, {self.dtype}')

    def _check_haplotype(self, haplotype, expected_shape):
        """Check haplotype has expected shape and dtype.
        Could extend to check values in {0,1,9}"""
        self._check_haplotype_dtype(haplotype)
        if haplotype.shape != expected_shape:
            raise ValueError('haplotype(s) has unexpected shape')

    def _sampled_library(self, indices):
        """Create a 'duplicated' HaplotypeLibrary consisting of specified indices only"""
        # Create HaplotypeLibrary object to return
        library = HaplotypeLibrary(self._n_loci)
        library.dtype = self.dtype
        library._haplotypes = self._haplotypes[indices]
        library._identifiers = self._identifiers[indices]
        library.freeze()
        return library


@jit(nopython=True)
def haplotype_from_indices(indices, haplotype_library):
    """Helper function that takes an array of indices (for each locus) that 'point' to rows
    in a haplotype library (NumPy array) and extracts the alleles from the corresponding haplotypes
    (in the library)
    Returns: a haplotype array of length n_loci"""

    n_loci = len(indices)
    haplotype = np.empty(n_loci, dtype=np.int8)
    for col_idx in range(n_loci):
        row_idx = indices[col_idx]
        haplotype[col_idx] = haplotype_library[row_idx, col_idx]
    return haplotype


@njit
def removeMissingValues(hap):
    for i in range(len(hap)) :
        if hap[i] == 9:
            hap[i] = random.randint(0, 1)

class ErrorLibrary(object):
    @profile
    def __init__(self, hap, haplotypes):
        self.hap = hap
        self.hapLib = haplotypes
        self.errors = jit_assessErrors(hap, self.hapLib)

    def getWindowValue(self, k):
        return jit_getWindowValue(self.errors, k, self.hap)

@jit(nopython=True)
def jit_getWindowValue(errors, k, hap) :
    window = np.full(errors.shape, 0, dtype = np.int8)
    nHaps, nLoci = errors.shape
    #Let me be silly here.
    for i in range(k+1):
        window[:,0] += errors[:,i]

    for i in range(1, nLoci):
        window[:,i] = window[:,i-1]
        if i > k:
            if hap[i-k-1] != 9:
                window[:,i] -= errors[:,i-k-1] #This is no longer in the window.
        if i < (nLoci-k):
            if hap[i+k] != 9:
                window[:,i] += errors[:,i+k] #This is now included in the window.

    return window

@jit(nopython=True)
def jit_assessErrors(hap, haps):
    errors = np.full(haps.shape, 0, dtype = np.int8)
    nHaps, nLoci = haps.shape
    for i in range(nLoci):
        if hap[i] != 9:
            if hap[i] == 0:
                errors[:, i] = haps[:,i]
            if hap[i] == 1:
                errors[:, i] = 1-haps[:,i]
    return errors

from collections import OrderedDict
class HaplotypeDict(object):
    def __init__(self):
        self.nHaps = 0
        self.haps = []
        self.tree = dict()
    # @profile
    def append(self, haplotype):
        byteVal = haplotype.tobytes()
        if byteVal in self.tree:
            return self.tree[byteVal]
        else:
            self.tree[byteVal] = self.nHaps
            self.haps.append(haplotype)
            self.nHaps += 1
            return self.nHaps -1
        return self.tree[byteVal]

    def get(self, index):
        return self.haps[index]

# hap = np.array([0, 0, 0, 0, 0, 0, 0, 0])

# hapLib = [np.array([1, 0, 0, 0, 0, 0, 1]),
#           np.array([0, 1, 0, 0, 0, 1, 0]),
#           np.array([0, 0, 1, 0, 1, 0, 0]),
#           np.array([0, 0, 0, 1, 0, 0, 0]),
#           np.array([0, 0, 1, 0, 1, 0, 0])]

# aa = ErrorLibrary(hap, hapLib)
# print(aa.errors)
# print(aa.getWindowValue(2))


# node_type = deferred_type()
# jit_randomBinary_spec = OrderedDict()
# jit_randomBinary_spec['array'] = int64[:]
# jit_randomBinary_spec['index'] = int64
# jit_randomBinary_spec['nItems'] = int64

# @jitclass(jit_randomBinary_spec)
# class jit_RandomBinary(object):
#     def __init__(self, nItems):
#         self.index = 0
#         self.nItems = nItems
#         self.array = np.random.randint(2, size = nItems)
#     def next():
#         self.index += 1
#         if self.index == self.nItems:
#             self.array = np.random.randint(2, size = nItems)
#             self.index = 0
#         return self.array[self.index]

# I Don't think this is used any more.
# def getCores(nLoci, lengths, offsets = [0]) :
#     nCores = len(lengths)*len(offsets)
#     startStop = []

#     for length in lengths:
#         for offset in offsets:
#             finished = False
#             if offset > 0:
#                 start = 0
#                 stop = min(offset, nLoci)
#                 startStop.append((start, stop))
#                 if stop == nLoci: finished = True
#             else:
#                 stop = 0
#             while not finished:
#                 start = stop
#                 stop = min(stop + length, nLoci)
#                 startStop.append((start, stop))
#                 if stop == nLoci: finished = True
#     return startStop

# from collections import OrderedDict
# class HaplotypeCount_dict(object):
#     def __init__(self):
#         self.nHaps = 0
#         self.haps = []
#         self.tree = OrderedDict()
#     # @profile
#     @profile
#     def append(self, haplotype, score = 1):
#         byteVal = haplotype.tobytes()
#         if byteVal in self.tree:
#             self.tree[byteVal] += score
#         else:
#             self.tree[byteVal] = score
#             self.haps.append(haplotype)
#             self.nHaps += 1

#     def getLargest(self):
#         #This is sloppy
#         vals = [value for key, value in self.tree.items()]

#         index = np.argmax(vals)
#         return self.haps[index]



from numba import jit
import numpy as np
import collections

def getGenotypesFromMaf(maf) :
    nLoci = len(maf)
    mafGenotypes = np.full((4, nLoci), .25, dtype = np.float32)

    mafGenotypes[0,:] = (1-maf)**2
    mafGenotypes[1,:] = maf*(1-maf)
    mafGenotypes[2,:] = (1-maf)*maf
    mafGenotypes[3,:] = maf**2

    return mafGenotypes

def getGenotypeProbabilities_ind(ind, args = None, log = False):
    if args is None:
        error = 0.01
        seqError = 0.001
        sexChromFlag = False
    else:
        error = args.error
        seqError = args.seqerror
        sexChromFlag = getattr(args, "sexchrom", False) and ind.sex == 0 #This is the sex chromosome and the individual is male.

    if ind.reads is not None:
        nLoci = len(ind.reads[0])
    if ind.genotypes is not None:
        nLoci = len(ind.genotypes)
    if not log:
        return getGenotypeProbabilities(nLoci, ind.genotypes, ind.reads, error, seqError, sexChromFlag)
    else:
        return getGenotypeProbabilities_log(nLoci, ind.genotypes, ind.reads, error, seqError, sexChromFlag)


def getGenotypeProbabilities(nLoci, genotypes, reads, error = 0.01, seqError = 0.001, useSexChrom=False):
    vals = np.full((4, nLoci), .25, dtype = np.float32)
    if type(error) is float:
        error = np.full(nLoci, error)
    if type(seqError) is float:
        seqError = np.full(nLoci, seqError)

    errorMat = generateErrorMat(error)

    if genotypes is not None:
        setGenoProbsFromGenotypes(genotypes, errorMat, vals)
        
    if reads is not None:
        seqError = seqError
        log1 = np.log(1-seqError)
        log2 = np.log(.5)
        loge = np.log(seqError)
        valSeq = np.array([log1*reads[0] + loge*reads[1],
                                log2*reads[0] + log2*reads[1],
                                log2*reads[0] + log2*reads[1],
                                log1*reads[1] + loge*reads[0]])
        maxVals = np.amax(valSeq, 0)
        valSeq = valSeq - maxVals
        valSeq = np.exp(valSeq)
        vals *= valSeq

    if useSexChrom:
        #Recode so we only care about the two homozygous states, but they are coded as 0, 1.
        vals[1,:] = vals[3,:]
        vals[2,:] = 0
        vals[3,:] = 0

    return vals/np.sum(vals,0)


def getGenotypeProbabilities_log(nLoci, genotypes, reads, error = 0.01, seqError = 0.001, useSexChrom=False):
    vals = np.full((4, nLoci), .25, dtype = np.float32)
    if type(error) is float:
        error = np.full(nLoci, error)
    if type(seqError) is float:
        seqError = np.full(nLoci, seqError)

    errorMat = generateErrorMat(error)

    if genotypes is not None:
        setGenoProbsFromGenotypes(genotypes, errorMat, vals)
    
    vals = np.log(vals)

    if reads is not None:
        log1 = np.log(1-seqError)
        log2 = np.log(.5)
        loge = np.log(seqError)

        ref_reads = reads[0]
        alt_reads = reads[1]

        val_seq = np.full((4, nLoci), 0, dtype = np.float32)
        val_seq[0,:] = log1*ref_reads + loge*alt_reads
        val_seq[1,:] = log2*ref_reads + log2*alt_reads
        val_seq[2,:] = log2*ref_reads + log2*alt_reads
        val_seq[3,:] = loge*ref_reads + log1*alt_reads

        vals += val_seq

    output = np.full((4, nLoci), 0, dtype = np.float32)
    apply_log_norm_1d(vals, output)
    return output

@jit(nopython=True, nogil = True)
def apply_log_norm_1d(vals, output):
    nLoci = vals.shape[-1]
    for i in range(nLoci):
        output[:,i] = log_norm_1D(vals[:, i])


@jit(nopython=True, nogil = True)
def log_norm_1D(mat):
    log_exp_sum = 0
    first = True
    maxVal = 100
    for a in range(4):
        if mat[a] > maxVal or first:
            maxVal = mat[a]
        if first:
            first = False

    for a in range(4):
        log_exp_sum += np.exp(mat[a] - maxVal)
    
    return mat - (np.log(log_exp_sum) + maxVal)


def set_from_genotype_probs(ind, geno_probs = None, calling_threshold = 0.1, set_genotypes = False, set_dosages = False, set_haplotypes = False) :

    # Check diploid geno_probs; not sure what to do for haploid except assume inbred?
    if geno_probs.shape[0] == 2:

        geno_probs = geno_probs/np.sum(geno_probs, axis = 0)
        called_values = call_genotype_probs(geno_probs, calling_threshold)

        # Assuming the individual is haploid

        if set_dosages:
            if ind.dosages is None:
                ind.dosages = called_values.dosages.copy()
            ind.dosages[:] = 2*called_values.dosages

        if set_genotypes:
            ind.genotypes[:] = 2*called_values.haplotypes
            ind.genotypes[called_values.haplotypes == 9] = 9 # Correctly set missing loci.

        if set_haplotypes:
            ind.haplotypes[0][:] = called_values.haplotypes
            ind.haplotypes[1][:] = called_values.haplotypes

    if geno_probs.shape[0] == 4:
        geno_probs = geno_probs/np.sum(geno_probs, axis = 0)
        called_values = call_genotype_probs(geno_probs, calling_threshold)

        if set_dosages:
            if ind.dosages is None:
                ind.dosages = called_values.dosages.copy()
            ind.dosages[:] = called_values.dosages

        if set_genotypes:
            ind.genotypes[:] = called_values.genotypes

        if set_haplotypes:
            ind.haplotypes[0][:] = called_values.haplotypes[0]
            ind.haplotypes[1][:] = called_values.haplotypes[1]


def call_genotype_probs(geno_probs, calling_threshold = 0.1) :

    if geno_probs.shape[0] == 2:
        # Haploid
        HaploidValues = collections.namedtuple("HaploidValues", ["haplotypes", "dosages"])
        dosages = geno_probs[1,:].copy()
        haplotypes = call_matrix(geno_probs, calling_threshold)

        return HaploidValues(dosages = dosages, haplotypes = haplotypes)

    if geno_probs.shape[0] == 4:
        # Diploid
        DiploidValues = collections.namedtuple("DiploidValues", ["genotypes", "haplotypes", "dosages"])

        dosages = geno_probs[1,:] + geno_probs[2,:] + 2*geno_probs[3,:]

        # Collapse the two heterozygous states into one.
        collapsed_hets = np.array([geno_probs[0,:], geno_probs[1,:] + geno_probs[2,:], geno_probs[3,:]], dtype=np.float32)
        genotypes = call_matrix(collapsed_hets, calling_threshold)

        # aa + aA, Aa + AA
        haplotype_0 = np.array([geno_probs[0,:] + geno_probs[1,:], geno_probs[2,:] + geno_probs[3,:]], dtype=np.float32)
        haplotype_1 = np.array([geno_probs[0,:] + geno_probs[2,:], geno_probs[1,:] + geno_probs[3,:]], dtype=np.float32)
        haplotypes = (call_matrix(haplotype_0, calling_threshold), call_matrix(haplotype_1, calling_threshold))
        
        return DiploidValues(dosages = dosages, haplotypes = haplotypes, genotypes = genotypes)


def call_matrix(matrix, threshold):
    called_genotypes = np.argmax(matrix, axis = 0)
    setMissing(called_genotypes, matrix, threshold)
    return called_genotypes.astype(np.int8)

@jit(nopython=True)
def setMissing(calledGenotypes, matrix, threshold) :
    nLoci = len(calledGenotypes)
    for i in range(nLoci):
        if matrix[calledGenotypes[i],i] < threshold:
            calledGenotypes[i] = 9


@jit(nopython=True)
def setGenoProbsFromGenotypes(genotypes, errorMat, vals):
    nLoci = len(genotypes)
    for i in range(nLoci) :
        if genotypes[i] != 9:
            vals[:, i] = errorMat[genotypes[i], :, i]
def generateErrorMat(error) :
    errorMat = np.array([[1-error, error/2, error/2, error/2], 
                            [error/2, 1-error, 1-error, error/2],
                            [error/2, error/2, error/2, 1-error]], dtype = np.float32)
    errorMat = errorMat/np.sum(errorMat, 1)[:,None]
    return errorMat


def generateSegregationXXChrom(partial=False, e= 1e-06) :
    paternalTransmission = np.array([ [1, 1, 0, 0],[0, 0, 1, 1]])
    maternalTransmission = np.array([ [1, 0, 1, 0],[0, 1, 0, 1]])

    fatherAlleleCoding = np.array([0, 0, 1, 1])
    motherAlleleCoding = np.array([0, 1, 0, 1])

    # !                  fm  fm  fm  fm 
    # !segregationOrder: pp, pm, mp, mm
    
    segregationTensor = np.zeros((4, 4, 4, 4))
    for segregation in range(4):
        #Change so that father always passes on the maternal allele?
        if(segregation == 0) :
            father = maternalTransmission
            mother = paternalTransmission
        if(segregation == 1) :
            father = maternalTransmission
            mother = maternalTransmission
        if(segregation == 2) :
            father = maternalTransmission
            mother = paternalTransmission
        if(segregation == 3) :
            father = maternalTransmission
            mother = maternalTransmission

        # !alleles: aa, aA, Aa, AA
        for allele in range(4) :
            segregationTensor[:, :, allele, segregation] = np.outer(father[fatherAlleleCoding[allele]], mother[motherAlleleCoding[allele]])

    segregationTensor = segregationTensor*(1-e) + e/4 #trace has 4 times as many elements as it should since it has 4 internal reps.
    segregationTensor = segregationTensor.astype(np.float32)

    return(segregationTensor)


def generateSegregationXYChrom(partial=False, e= 1e-06) :
    paternalTransmission = np.array([ [1, 1, 0, 0],[0, 0, 1, 1]])
    maternalTransmission = np.array([ [1, 0, 1, 0],[0, 1, 0, 1]])

    motherAlleleCoding = np.array([0, 1, 0, 1])

    # !                  fm  fm  fm  fm 
    # !segregationOrder: pp, pm, mp, mm
    #They don't get anything from the father -- father is always 0
    segregationTensor = np.zeros((4, 4, 4, 4))
    for segregation in range(4):
        if(segregation == 0) :
            mother = paternalTransmission
        if(segregation == 1) :
            mother = maternalTransmission
        if(segregation == 2) :
            mother = paternalTransmission
        if(segregation == 3) :
            mother = maternalTransmission

        # !alleles: aa, aA, Aa, AA
        for allele in range(4) :
            for fatherAllele in range(4):
                segregationTensor[fatherAllele, :, allele, segregation] = mother[motherAlleleCoding[allele]]

    segregationTensor = segregationTensor*(1-e) + e/4 #trace has 4 times as many elements as it should since it has 4 internal reps.
    segregationTensor = segregationTensor.astype(np.float32)

    return(segregationTensor)



def generateSegregation(partial=False, e= 1e-06) :
    paternalTransmission = np.array([ [1, 1, 0, 0],[0, 0, 1, 1]])
    maternalTransmission = np.array([ [1, 0, 1, 0],[0, 1, 0, 1]])

    fatherAlleleCoding = np.array([0, 0, 1, 1])
    motherAlleleCoding = np.array([0, 1, 0, 1])

    # !                  fm  fm  fm  fm 
    # !segregationOrder: pp, pm, mp, mm
    
    segregationTensor = np.zeros((4, 4, 4, 4))
    for segregation in range(4):
        if(segregation == 0) :
            father = paternalTransmission
            mother = paternalTransmission
        if(segregation == 1) :
            father = paternalTransmission
            mother = maternalTransmission
        if(segregation == 2) :
            father = maternalTransmission
            mother = paternalTransmission
        if(segregation == 3) :
            father = maternalTransmission
            mother = maternalTransmission

        # !alleles: aa, aA, Aa, AA
        for allele in range(4) :
            segregationTensor[:, :, allele, segregation] = np.outer(father[fatherAlleleCoding[allele]], mother[motherAlleleCoding[allele]])

    if partial : segregationTensor = np.mean(segregationTensor, 3)

    segregationTensor = segregationTensor*(1-e) + e/4 #trace has 4 times as many elements as it should since it has 4 internal reps.
    segregationTensor = segregationTensor.astype(np.float32)
    return(segregationTensor)

# def generateErrorMat(error) :
#     # errorMat = np.array([[1-error*3/4, error/4, error/4, error/4], 
#     #                         [error/4, .5-error/4, .5-error/4, error/4],
#     #                         [error/4, error/4, error/4, 1-error*3/4]], dtype = np.float32)
#     errorMat = np.array([[1-error*2/3, error/3, error/3, error/3], 
#                             [error/3, 1-error*2/3, 1-error*2/3, error/3],
#                             [error/3, error/3, error/3, 1-error*2/3]], dtype = np.float32)
#     errorMat = errorMat/np.sum(errorMat, 1)[:,None]
#     return errorMat

## Not sure if below is ever used. 
# def generateTransmission(error) :
#     paternalTransmission = np.array([ [1-error, 1.-error, error, error],
#                                       [error, error, 1-error, 1-error]])
    
#     maternalTransmission = np.array([ [1.-error, error, 1.-error, error],
#                                         [error, 1-error, error, 1-error]] )
#     segregationTransmissionMatrix = np.zeros((4,4))
#     segregationTransmissionMatrix[0,:] = paternalTransmission[0,:]
#     segregationTransmissionMatrix[1,:] = paternalTransmission[0,:]
#     segregationTransmissionMatrix[2,:] = paternalTransmission[1,:]
#     segregationTransmissionMatrix[3,:] = paternalTransmission[1,:]

#     segregationTransmissionMatrix[:,0] = segregationTransmissionMatrix[:,0] * maternalTransmission[0,:]
#     segregationTransmissionMatrix[:,1] = segregationTransmissionMatrix[:,1] * maternalTransmission[1,:]
#     segregationTransmissionMatrix[:,2] = segregationTransmissionMatrix[:,2] * maternalTransmission[0,:]
#     segregationTransmissionMatrix[:,3] = segregationTransmissionMatrix[:,3] * maternalTransmission[1,:]

#     segregationTransmissionMatrix = segregationTransmissionMatrix.astype(np.float32)
#     return(segregationTransmissionMatrix)



import random
import numpy as np
import numba
from numba import njit, jit, int8, int32,int64, boolean, deferred_type, optional, float32
from numba.experimental import jitclass
from collections import OrderedDict

try:
    profile
except:
    def profile(x): 
        return x

#####################################
###                               ###
###       Burrows Wheeler         ###
###                               ###
#####################################

class BurrowsWheelerLibrary():
    def __init__(self, haplotypeList):
        self.library = createBWLibrary(np.array(haplotypeList))
        self.hapList = haplotypeList
        self.nHaps = len(haplotypeList)
    def getHaplotypeMatches(self, haplotype, start, stop):
        nHaps, hapIndexes = getConsistentHaplotypes(self.library, haplotype, start, stop)
        haps = [(self.hapList[hapIndexes[index, 0]], hapIndexes[index, 1]) for index in range(nHaps)]
        return haps
    @profile
    def getBestHaplotype(self, weights, start, stop):
        index = getHaplotypesPlusWeights(self.library, weights, start, stop)

        return self.hapList[index][start:stop]

jit_BurrowsWheelerLibrary_spec = OrderedDict()
jit_BurrowsWheelerLibrary_spec['a'] = int64[:,:]
jit_BurrowsWheelerLibrary_spec['d'] = int64[:,:]
jit_BurrowsWheelerLibrary_spec['zeroOccPrev'] = int64[:,:]
jit_BurrowsWheelerLibrary_spec['nZeros'] = int64[:]
jit_BurrowsWheelerLibrary_spec['haps'] = int8[:,:]

@jitclass(jit_BurrowsWheelerLibrary_spec)
class jit_BurrowsWheelerLibrary():
    def __init__(self, a, d, nZeros, zeroOccPrev, haps):
        self.a = a
        self.d = d
        self.nZeros = nZeros
        self.zeroOccPrev = zeroOccPrev
        self.haps = haps

    def getValues(self):
        return (self.a, self.d, self.nZeros, self.zeroOccPrev, self.haps)


    def update_state(self, state, index):
        pass

    def get_null_state(self, value, index):
        if value == 0:
            lowerR = 0
            upperR = nZeros[stop-1]

        if value == 1:
            lowerR = nZeros[stop-1]
            upperR = nHaps

        pass

@njit
def createBWLibrary(haps):
    
    #Definitions.
    # haps : a list of haplotypes
    # a : an ordering of haps in lexographic order.
    # d : Number of loci of a[i,j+k] == a[i,-1, j+k]


    nHaps = haps.shape[0]
    nLoci = haps.shape[1]
    a = np.full(haps.shape, 0, dtype = np.int64)
    d = np.full(haps.shape, 0, dtype = np.int64)

    nZerosArray = np.full(nLoci, 0, dtype = np.int64)

    zeros = np.full(nHaps, 0, dtype = np.int64)
    ones = np.full(nHaps, 0, dtype = np.int64)
    dZeros = np.full(nHaps, 0, dtype = np.int64)
    dOnes = np.full(nHaps, 0, dtype = np.int64)
    
    nZeros = 0
    nOnes = 0
    for j in range(nHaps):
        if haps[j, nLoci-1] == 0:
            zeros[nZeros] = j
            if nZeros == 0:
                dZeros[nZeros] = 0
            else:
                dZeros[nZeros] = 1
            nZeros += 1
        else:
            ones[nOnes] = j    
            if nOnes == 0:
                dOnes[nOnes] = 0
            else:
                dOnes[nOnes] = 1
            nOnes += 1
    if nZeros > 0:
        a[0:nZeros, nLoci-1] = zeros[0:nZeros]
        d[0:nZeros, nLoci-1] = dZeros[0:nZeros]

    if nOnes > 0:
        a[nZeros:nHaps, nLoci-1] = ones[0:nOnes]
        d[nZeros:nHaps, nLoci-1] = dOnes[0:nOnes]

    nZerosArray[nLoci-1] = nZeros

    for i in range(nLoci-2, -1, -1) :
        zeros = np.full(nHaps, 0, dtype = np.int64)
        ones = np.full(nHaps, 0, dtype = np.int64)
        dZeros = np.full(nHaps, 0, dtype = np.int64)
        dOnes = np.full(nHaps, 0, dtype = np.int64)
    
        nZeros = 0
        nOnes = 0

        dZerosTmp = -1 #This is a hack.
        dOnesTmp = -1

        for j in range(nHaps) :

            dZerosTmp = min(dZerosTmp, d[j,i+1])
            dOnesTmp = min(dOnesTmp, d[j,i+1])
            if haps[a[j, i+1], i] == 0:
                zeros[nZeros] = a[j, i+1]
                dZeros[nZeros] = dZerosTmp + 1
                nZeros += 1
                dZerosTmp = nLoci
            else:
                ones[nOnes] = a[j, i+1]
                dOnes[nOnes] = dOnesTmp + 1
                nOnes += 1
                dOnesTmp = nLoci


        if nZeros > 0:
            a[0:nZeros, i] = zeros[0:nZeros]
            d[0:nZeros, i] = dZeros[0:nZeros]

        if nOnes > 0:
            a[nZeros:nHaps, i] = ones[0:nOnes]
            d[nZeros:nHaps, i] = dOnes[0:nOnes]
        nZerosArray[i] = nZeros


    #I'm going to be a wee bit sloppy in creating zeroOccPrev
    #Not defined at 0 so start at 1.
    zeroOccPrev = np.full(haps.shape, 0, dtype = np.int64)

    for i in range(1, nLoci):
        count = 0
        for j in range(0, nHaps):
            if haps[a[j, i], i-1] == 0:
                count += 1
            zeroOccPrev[j, i] = count


    library = jit_BurrowsWheelerLibrary(a, d, nZerosArray, zeroOccPrev, haps)
    return library

@jit(nopython=True, nogil=True)
def getConsistentHaplotypes(bwLibrary, hap, start, stop):
    a, d, nZeros, zeroOccPrev, haps = bwLibrary.getValues()
    nHaps = a.shape[0]
    nLoci = a.shape[1]


    intervals = np.full((nHaps, 2), 0, dtype = np.int64)
    intervals_new = np.full((nHaps, 2), 0, dtype = np.int64)
    
    nIntervals = 0
    nIntervals_new = 0
    
    #Haps go from 0 to nHaps-1. Loci go from start to stop-1 (inclusive).
    #The first hap with one is nZeros. The last hap with zero is nZeros -1.
    #Last loci is stop -1
    #These are split out because they represent *distinct* haplotypes.
    #Maybe could do this with tuple and list append but *shrug*.

    if hap[stop-1] == 0 or hap[stop-1] == 9:
        lowerR = 0
        upperR = nZeros[stop-1]
        if upperR >= lowerR:
            intervals[nIntervals, 0] = lowerR
            intervals[nIntervals, 1] = upperR
            nIntervals += 1
    
    if hap[stop-1] == 1 or hap[stop-1] == 9:
        lowerR = nZeros[stop-1]
        upperR = nHaps
        if upperR >= lowerR:
            intervals[nIntervals, 0] = lowerR
            intervals[nIntervals, 1] = upperR
            nIntervals += 1

    #Python indexing is annoying.
    #Exclude stop and stop-1, include start.
    #Intervals are over haplotypes.
    for i in range(stop-2, start-1, -1):
        # print(intervals[0:nIntervals,:])

        nIntervals_new = 0

        #Doing it on interval seems to make marginally more sense.
        for interval in range(nIntervals):

            int_start = intervals[interval, 0]
            int_end = intervals[interval, 1]

            if hap[i] == 0 or hap[i] == 9:
                if int_start == 0:
                    lowerR = 0
                else:
                    lowerR = zeroOccPrev[int_start-1, i+1] 
                upperR = zeroOccPrev[int_end-1, i+1] #Number of zeros in the region. 
                if upperR > lowerR: #Needs to be greater than. Regions no longer inclusive.
                    # print("Added 0:", int_start, int_end, "->>", lowerR, upperR)
                    intervals_new[nIntervals_new, 0] = lowerR
                    intervals_new[nIntervals_new, 1] = upperR
                    nIntervals_new += 1
            if hap[i] == 1 or hap[i] == 9:

                # of ones between 0 and k (k inclusive) is k+1 - number of zeros.

                if int_start == 0:
                    lowerR = nZeros[i]
                else:
                    lowerR = nZeros[i] + (int_start - zeroOccPrev[int_start-1, i+1]) 
                upperR = nZeros[i] + (int_end - zeroOccPrev[int_end-1, i+1])
                if upperR > lowerR:
                    # print("Added 1:", int_start, int_end, "->>", lowerR, upperR)
                    intervals_new[nIntervals_new, 0] = lowerR
                    intervals_new[nIntervals_new, 1] = upperR
                    nIntervals_new += 1
                # else:
                    # print(i, "interval rejected:", int_start, int_end, "->", upperR, lowerR)
        #This is basically intervals = intervals_new
        for j in range(nIntervals_new):
            intervals[j, 0] = intervals_new[j, 0]
            intervals[j, 1] = intervals_new[j, 1]
        nIntervals = nIntervals_new
        # print("Finished", i, "->", nIntervals)
    # print(intervals[0:nIntervals,:])

    hapIndexes = np.full((nHaps, 2), 0, dtype = np.int64)
    nHapsAssigned = 0
    for i in range(nIntervals):

        int_start = intervals[i, 0]
        int_end = intervals[i, 1]


        hapIndexes[nHapsAssigned, 0] = a[int_start,start]
        hapIndexes[nHapsAssigned, 1] = int_end - int_start
        nHapsAssigned +=1

    return (nHapsAssigned, hapIndexes)

# def printSortAt(loci, library):
#     a, d, nZeros, zeroOccPrev, haps = library.getValues()
#     vals = haps[a[:,loci],:]
#     for i in range(vals.shape[0]):
#         print(i, ' '.join(map(str, vals[i,:])) )
    
#     # print(zeroOccPrev[:,:])

# hapLib = [np.array([1, 0, 0, 0, 0, 0, 1], dtype = np.int8),
#           np.array([0, 1, 0, 0, 0, 1, 0], dtype = np.int8),
#           np.array([0, 1, 0, 0, 0, 1, 0], dtype = np.int8),
#           np.array([0, 1, 0, 0, 0, 1, 0], dtype = np.int8),
#           np.array([0, 0, 1, 0, 1, 0, 0], dtype = np.int8),
#           np.array([1, 1, 1, 0, 0, 0, 0], dtype = np.int8),
#           np.array([0, 0, 1, 0, 1, 0, 0], dtype = np.int8),
#           np.array([1, 1, 1, 0, 1, 0, 0], dtype = np.int8),
#           np.array([0, 0, 0, 1, 0, 0, 0], dtype = np.int8),
#           np.array([0, 1, 1, 1, 0, 0, 0], dtype = np.int8),
#           np.array([0, 1, 1, 1, 0, 0, 0], dtype = np.int8),
#           np.array([1, 1, 0, 1, 0, 0, 0], dtype = np.int8),
#           np.array([0, 0, 1, 0, 1, 0, 0], dtype = np.int8),
#           np.array([0, 0, 1, 0, 1, 0, 0], dtype = np.int8),
#           np.array([0, 0, 1, 0, 1, 0, 0], dtype = np.int8),
#           np.array([0, 0, 1, 0, 1, 0, 0], dtype = np.int8),
#           np.array([0, 0, 1, 0, 1, 0, 0], dtype = np.int8)]

# bwlib = BurrowsWheelerLibrary(hapLib)

# # printSortAt(0, bwlib.library)
# printSortAt(6, bwlib.library); print("")
# printSortAt(5, bwlib.library); print("")
# printSortAt(4, bwlib.library); print("")
# printSortAt(3, bwlib.library); print("")
# printSortAt(2, bwlib.library); print("")
# printSortAt(1, bwlib.library); print("")
# printSortAt(0, bwlib.library); print("")

# # print(bwlib.getHaplotypeMatches(haplotype = np.array([0, 0, 0], dtype = np.int8), start = 0, stop = 3))
# tmp = (bwlib.getHaplotypeMatches(haplotype = np.array([9, 9, 9, 9, 9, 9, 9], dtype = np.int8), start = 0, stop = 7))
# for key, value in tmp:
#     print(key, value)

@njit
def getConsistentHaplotypesBruteForce(bwLibrary, hap, start, stop):
    hap = hap[start:stop]

    a, d, nZeros, zeroOccPrev, haps = bwLibrary.getValues()
    recodedHaps = haps[:, start:stop]
    
    nHaps = recodedHaps.shape[0]
    nLoci = recodedHaps.shape[1]

    consistent = np.full(nHaps, 0, dtype = np.int64)

    #Last correct index
    lastIndex = -1
    for j in range(nHaps):

        #Otherwise, we have not enough information and need to search.
        add = True
        for i in range(nLoci):
            if hap[i] != 9 :
                if recodedHaps[j, i] != hap[i]:
                    add = False
        if add:
            consistent[j] = 1
    
    hapIndexes = np.full((nHaps, 2), 0, dtype = np.int64)
    nHapsAssigned = 0
    for i in range(nHaps):
        if consistent[i] > 0:
            # hapIndexes[nHapsAssigned, 0] = a[i,start]
            hapIndexes[nHapsAssigned, 0] = i
            hapIndexes[nHapsAssigned, 1] = consistent[i]
            nHapsAssigned +=1

    return (nHapsAssigned, hapIndexes)
@njit
def getHaplotypesPlusWeights(bwLibrary, weights, start, stop):
    #Weights are weights to the original haplotypes (haps)

    a, d, nZeros, zeroOccPrev, haps = bwLibrary.getValues()
    recodedWeights = weights[a[:, start]]

    nHaps = d.shape[0]
    nLoci = stop - start
    bestLoci = -1
    bestWeight = -1
    currentLoci = 0
    currentWeight = 0

    for j in range(nHaps):
        #Will need to double check this. This code will be slow!

        if d[j, start] < nLoci:
            #Process the last haplotype before moving on.
            if currentWeight > bestWeight :
                bestLoci = currentLoci
                bestWeight = currentWeight
            
            currentLoci = j
            currentWeight = 0

        currentWeight += recodedWeights[j]

    #Make sure we check the last haplotype.
    if currentWeight > bestWeight :
        bestLoci = currentLoci
        bestWeight = currentWeight


    #REMEMBER TO RECODE
    return a[bestLoci, start]

#Older version that doesn't use all of the meta data we have.
# @jit(nopython=True, nogil=True)
# def getConsistentHaplotypes(bwLibrary, hap, start, stop):
#     a, d, nZeros, zeroOccPrev, haps = bwLibrary.getValues()
#     hap = hap[start:stop]
#     recodedHaps = haps[a[:, start], start:stop]
    
#     nHaps = recodedHaps.shape[0]
#     nLoci = recodedHaps.shape[1]
#     consistent = np.full(nHaps, 0, dtype = np.int64)

#     lastCorrect = -1
#     firstError = nLoci + 1

#     #Last correct index
#     lastIndex = -1
#     for j in range(nHaps):
#         #Basically, we know how much overlap there was with the previous haplotype.
#         #We can use that to get a better estimate of where this one will be correct.

#         #By definition, all of 0 -> d[j, start]-1 inclusive is the same.
#         #All of 0 -> lastCorrect (inclusive) is correct.
#         #First error is the position of the first error. If firstError < nLoci, this is a problem. (nLoci is out of our bounds)
        
#         lastCorrect = min(lastCorrect, d[j, start]-1)
#         if firstError > d[j,start]-1: firstError = nLoci
        
#         #Two elif statements.
#         #First can we spot an error?
#         #Second if no error's exist, are we sure that the entire haplotype is right.
#         if firstError < nLoci: #or equal?
#             consistent[j] = 0
#             lastIndex = -1
#         #nLoci is the last position we care about (nLoci is out of our bounds). If nLoci is correct, then we're good).
#         elif lastCorrect >= nLoci-1:
#             #Adding in some short circuit code to prevent duplication
#             # lastIndex = -1 ###THIS LINE TOGGLES DUPLICATION PREVENTION
#             if lastIndex != -1:
#                 consistent[lastIndex] += 1
#             else:
#                 consistent[j] = 1
#                 lastIndex = j
#         else:
#             #Otherwise, we have not enough information and need to search.
#             #Last correct is the last known correct loci.
#             stopIters = False
#             i = lastCorrect
#             while not stopIters:
#                 i = i+1 #We know that the last correct loci is correct. So we're safe to start at last correct +1 
#                 if hap[i] != 9 and recodedHaps[j, i] != hap[i]:
#                     lastCorrect = i-1
#                     firstError = i
#                     stopIters = True
#                 elif i == nLoci-1:
#                     stopIters = True
#                     lastCorrect = nLoci-1
#                     firstError = nLoci 

#             if firstError < nLoci: 
#                 consistent[j] = 0
#                 lastIndex = -1

#             elif lastCorrect >= nLoci-1: #This will probably be nLoci-1 since that is where our search stops.
#                 consistent[j] = 1
#                 lastIndex = j
    
#     hapIndexes = np.full((nHaps, 2), 0, dtype = np.int64)
#     nHapsAssigned = 0
#     for i in range(nHaps):
#         if consistent[i] > 0:
#             hapIndexes[nHapsAssigned, 0] = a[i,start]
#             hapIndexes[nHapsAssigned, 1] = consistent[i]
#             nHapsAssigned +=1

#     return (nHapsAssigned, hapIndexes)



from numba import jit
import numpy as np
from . import NumbaUtils
from . HaplotypeLibrary import haplotype_from_indices

def haploidHMM(individual, source_haplotypes, error, recombination_rate, threshold=0.9, calling_method='dosages'):

    target_haplotype = individual.haplotypes
    n_loci = len(target_haplotype)

    # !!!! May need to cast the source Haplotypes to a matrix. #May also want to handle the probabilistic case.
    if type(source_haplotypes) is list:
        source_haplotypes = np.array(source_haplotypes)

    # Expand error and recombinationRate to arrays as may need to have
    # marker specific error/recombination rates.
    if type(error) is float:
        error = np.full(n_loci, error, dtype=np.float32)
    if type(recombination_rate) is float:
        recombination_rate = np.full(n_loci, recombination_rate, dtype=np.float32)

    # Construct penetrance values (point estimates)
    point_estimates = getHaploidPointEstimates(target_haplotype, source_haplotypes, error)

    # Run forward-backward algorithm on penetrance values
    # Note: don't need these probabilites if using the sample method
    if calling_method != 'sample':
        total_probs = haploidForwardBackward(point_estimates, recombination_rate)

    # Handle the different calling methods
    if calling_method == 'callhaps':
        # Call haplotypes
        called_haps = haploidCallHaps(total_probs, threshold)
        # Call genotypes
        called_genotypes = getHaploidGenotypes(called_haps, source_haplotypes)
        return called_genotypes
    if calling_method == 'dosages':
        dosages = getHaploidDosages(total_probs, source_haplotypes)
        individual.dosages = dosages
    if calling_method == 'sample':
        haplotype = getHaploidSample(point_estimates, recombination_rate, source_haplotypes)
        individual.imputed_haplotypes = haplotype
    if calling_method == 'Viterbi':
        haplotype = get_viterbi(point_estimates, recombination_rate, source_haplotypes)
        individual.imputed_haplotypes = haplotype


@jit(nopython=True, nogil=True)
def getHaploidDosages(hap_est, source_haplotypes):
    """Calculate dosages for a single haplotype"""
    n_loci, n_haps = hap_est.shape
    dosages = np.zeros(n_loci, dtype=np.float32)
    for i in range(n_loci):
        for j in range(n_haps):
            dosages[i] += source_haplotypes[j, i] * hap_est[i, j]
    return dosages


@jit(nopython=True, nogil=True)
def getHaploidSample(point_estimates, recombination_rate, source_haps):
    """Sample a haplotype"""
    forward_probs = haploidForward(point_estimates, recombination_rate)
    haplotype = haploidSampleHaplotype(forward_probs, source_haps, recombination_rate)
    return haplotype


@jit(nopython=True, nogil=True)
def get_viterbi(point_estimates, recombination_rate, haplotype_library):
    """Get most likely haplotype using the Viterbi algorithm"""
    forward_probs = haploidForward(point_estimates, recombination_rate)
    indices = haploid_viterbi(forward_probs, recombination_rate)
    return haplotype_from_indices(indices, haplotype_library)


@jit(nopython=True)
def haploidCallHaps(hapEst, threshold):
    nHaps, nLoci = hapEst.shape
    calledHaps = np.full(nLoci, -1, dtype=np.int64)  # These are haplotype ids. -1 is missing.
    for i in range(nLoci):
        maxVal = -1
        maxLoc = -1
        for j in range(nHaps):
            if hapEst[j, i] > threshold and hapEst[j, i] > maxVal:
                maxLoc = j
                maxVal = hapEst[j, i]
        calledHaps[i] = maxLoc
    return calledHaps


@jit(nopython=True)
def getHaploidGenotypes(calledHaps, sourceHaplotypes):
    nHaps, nLoci = sourceHaplotypes.shape
    calledGenotypes = np.full(nLoci, 9, dtype=np.int8) # These are haplotype ids. -1 is missing.
    for i in range(nLoci):
        if calledHaps[i] != -1:
            calledGenotypes[i] = sourceHaplotypes[calledHaps[i], i]
    return calledGenotypes


@jit(nopython=True, nogil=True)
def getHaploidPointEstimates(targetHaplotype, sourceHaplotypes, error):
    nHaps, nLoci = sourceHaplotypes.shape
    pointMat = np.full((nLoci, nHaps), 1, dtype=np.float32)

    for i in range(nLoci):
        if targetHaplotype[i] != 9:
            for j in range(nHaps):
                if targetHaplotype[i] == sourceHaplotypes[j, i]:
                    pointMat[i, j] = 1 - error[i]
                else:
                    pointMat[i, j] = error[i]
    return pointMat


@jit(nopython=True, nogil=True)
def haploidTransformProbs(previous, new, estimate, point_estimate, recombination_rate):
    """Transforms a probability distribution (over haplotypes, at a single locus)
    to a probability distribution at the next locus by accounting for emission probabilities
    (point_estimates) and transition probabilities (recombination_rate)
    This is a core step in the forward and backward algorithms

    point_estimates     emission probabilities - (1D NumPy array)
    recombination_rate  recombination rate at this locus - (scalar)
    previous            probability distribution over haplotypes (hidden states) at the *previous* locus
    estimate            newly calculated probability distribution over haplotypes at *this* locus
    new                 intermediate probability distribution (passed in to this function for speed)

    Note: previous and estimate are updated by this function
    """
    n_haps = len(previous)

    # Get estimate at this locus and normalize
    new[:] = previous * point_estimate
    new /= np.sum(new)

    # Account for recombination rate
    e = recombination_rate
    e1 = 1-recombination_rate
    for j in range(n_haps):
        new[j] = new[j]*e1 + e/n_haps

    # Update distributions (in place)
    for j in range(n_haps):
        estimate[j] *= new[j]
        previous[j] = new[j]


@jit(nopython=True, nogil=True)
def haploidOneSample(forward_probs, recombination_rate):
    """Sample one haplotype (an individual) from the forward and backward probability distributions
    Returns two arrays:
      sample_indices   array of indices of haplotypes in the haplotype library at each locus
                       e.g. an individual composed of haplotypes 13 and 42 with 8 loci:
                       [42, 42, 42, 42, 42, 13, 13, 13]
    A description of the sampling process would be nice here..."""

    est = forward_probs.copy()  # copy so that forward_probs is not modified
    n_loci, n_haps = forward_probs.shape
    prev = np.ones(n_haps, dtype=np.float32)
    new = np.empty(n_haps, dtype=np.float32)

    # Sampled probability distribution at one locus
    sampled_probs = np.empty(n_haps, dtype=np.float32)
    sample_indices = np.empty(n_loci, dtype=np.int64)

    # Backwards algorithm
    for i in range(n_loci-2, -1, -1): # zero indexed then minus one since we skip the boundary
        # Sample at this locus
        j = NumbaUtils.multinomial_sample(pvals=est[i+1, :])
        sampled_probs[:] = 0
        sampled_probs[j] = 1
        sample_indices[i+1] = j

        # Get estimate at this locus using the *sampled* distribution
        # (instead of the point estimates/emission probabilities)
        haploidTransformProbs(prev, new, est[i, :], sampled_probs, recombination_rate[i+1])
        # No need to normalise at this locus as multinomial_sample()
        # handles un-normalized probabilities

    # Last sample (at the first locus)
    j = NumbaUtils.multinomial_sample(pvals=est[0, :])
    sample_indices[0] = j

    return sample_indices


@jit(nopython=True, nogil=True)
def haploid_viterbi(forward_probs, recombination_rate):
    """Find the most likely haplotype according to the The Viterbi algorithm
    Returns:
      indices   array of indices of haplotypes in the haplotype library at each locus
                e.g. an individual composed of haplotypes 13 and 42 with 8 loci:
                [42, 42, 42, 42, 42, 13, 13, 13]"""

    est = forward_probs.copy()  # copy so that forward_probs is not modified
    n_loci, n_haps = forward_probs.shape
    prev = np.ones(n_haps, dtype=np.float32)
    new = np.empty(n_haps, dtype=np.float32)

    # Most likely probability distribution at one locus
    sampled_probs = np.empty(n_haps, dtype=np.float32)
    indices = np.empty(n_loci, dtype=np.int64)

    # Backwards algorithm
    for i in range(n_loci-2, -1, -1): # zero indexed then minus one since we skip the boundary
        # Choose the most likely state (i.e. max probability) at this locus
        j = np.argmax(est[i+1, :])

        sampled_probs[:] = 0
        sampled_probs[j] = 1
        indices[i+1] = j

        # Get estimate at this locus using the most likely distribution
        # (instead of the point estimates/emission probabilities)
        haploidTransformProbs(prev, new, est[i, :], sampled_probs, recombination_rate[i+1])
        # No need to normalise at this locus as argmax() does not depend on normalisation

    # Most likely state at the first locus
    j = np.argmax(est[0, :])
    indices[0] = j

    return indices


@jit(nopython=True, nogil=True)
def haploidSampleHaplotype(forward_probs, haplotype_library, recombination_rate):
    """Sample one haplotype (an individual) from the forward and backward probability distributions
    Returns: a sampled haploytpe of length n_loci"""
    indices = haploidOneSample(forward_probs, recombination_rate)
    return haplotype_from_indices(indices, haplotype_library)


@jit(nopython=True, nogil=True)
def haploidForward(point_estimate, recombination_rate):
    """Calculate (unnomalized) forward probabilities"""

    n_loci, n_haps = point_estimate.shape
    est = point_estimate.copy()
    prev = np.ones(n_haps, dtype=np.float32)
    new = np.empty(n_haps, dtype=np.float32)

    for i in range(1, n_loci):
        # Update estimates at this locus
        haploidTransformProbs(prev, new, est[i, :], point_estimate[i-1, :], recombination_rate[i])

    return est


@jit(nopython=True)
def haploidBackward(point_estimate, recombination_rate):
    """Calculate (unnomalized) backward probabilities"""

    n_loci, n_haps = point_estimate.shape
    est = np.ones_like(point_estimate, dtype=np.float32)
    prev = np.ones(n_haps, dtype=np.float32)
    new = np.empty(n_haps, dtype=np.float32)

    for i in range(n_loci-2, -1, -1):  # zero indexed then minus one since we skip the boundary
        # Update estimates at this locus
        haploidTransformProbs(prev, new, est[i, :], point_estimate[i+1, :], recombination_rate[i+1])

    return est


@jit(nopython=True)
def haploidForwardBackward(point_estimate, recombination_rate):
    """Calculate normalized state probabilities at each loci using the forward-backward algorithm"""

    est = (haploidForward(point_estimate, recombination_rate) *
           haploidBackward(point_estimate, recombination_rate))

    # Return normalized probabilities
    n_loci = point_estimate.shape[0]
    for i in range(n_loci):
        est[i, :] /= np.sum(est[i, :])

    return est



from setuptools import setup, Extension, find_packages

setup(
    name="AlphaPeel",
    version="1.1.1",
    author="Andrew Whalen",
    author_email="awhalen@roslin.ed.ac.uk",
    description="Multilocus and hybrid peeling",
    long_description="This is a package for performing multi-locus and hybrid peeling in large pedigree populations. For more information see Hybrid peeling for fast and accurate calling, phasing, and imputation with sequence data of any coverage in pedigrees (Whalen et al, 2018; doi: 10.1186/s12711-018-0438-2) ",
    long_description_content_type="text/markdown",
    packages=['tinypeel', 'tinypeel.tinyhouse','tinypeel.Peeling'],
    package_dir={'': 'src'},
    license="MIT license",

    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    entry_points = {
    'console_scripts': [
        'AlphaPeel=tinypeel.tinypeel:main',
        ],
    },
    install_requires=[
        'numpy>=1.19',
        'numba>=0.50.0'
    ]
)



from numba import jit
import numpy as np
from . import NumbaUtils
from . import ProbMath

class HaploidMarkovModel :
    def __init__(self, n_loci, error, recombination_rate = None):

        self.update_paramaters(n_loci, error, recombination_rate)

        self.directional_smoothing = self.create_directional_smoothing()
        self.apply_smoothing = self.create_apply_smoothing()
        self.apply_viterbi = self.create_viterbi_algorithm()
        self.apply_sampling = self.create_sampling_algorithm(NumbaUtils.multinomial_sample)


    def update_paramaters(self, n_loci, error, recombination_rate = None):

        self.n_loci = n_loci

        if type(error) is float:
            self.error = np.full(n_loci, error, dtype=np.float32)
        else:
            self.error = error

        if recombination_rate is None:
            recombination_rate = 1.0/n_loci

        if type(recombination_rate) is float:
            self.recombination_rate = np.full(n_loci, recombination_rate, dtype=np.float32)
        else:
            self.recombination_rate = recombination_rate


    def get_mask(self, called_haplotypes):
        return np.all(called_haplotypes != 9, axis = 0)

    # def get_run_option(default_arg, alternative_arg):
    #     # Return the default arg as true if it is supplied, otherwise return the alternative arg.
    #     if default_arg is not None:
    #         if alternative_arg is not None:
    #             if default_arg and alternative_arg:
    #                 print("Both arguments are true, returning default")
    #             if not default_arg and not alternative_arg:
    #                 print("Both arguments are false, returning default")
    #         return default_arg
    #     else:
    #         if alternative_arg is None:
    #             return True
    #         else:
    #             return not alternative_arg


    def run_HMM(self, point_estimates = None, algorithm = "marginalize", **kwargs):
        
        # return_called_values = get_run_option(return_called_values, return_genotype_probabilities)

        if point_estimates is None:
            point_estimates = self.get_point_estimates(**kwargs)
        
        if algorithm == "marginalize":
            total_probs = self.apply_smoothing(point_estimates, self.recombination_rate)
            genotype_probabilities = self.calculate_genotype_probabilities(total_probs, **kwargs)

        elif algorithm == "viterbi":
            total_probs = self.apply_viterbi(point_estimates, self.recombination_rate)
            genotype_probabilities = self.calculate_genotype_probabilities(total_probs, **kwargs)
        
        elif algorithm == "sample":
            total_probs = self.apply_sampling(point_estimates, self.recombination_rate)
            genotype_probabilities = self.calculate_genotype_probabilities(total_probs, **kwargs)

        else:
            print(f"Valid alrogithm option not given: {alrogithm}")

        return genotype_probabilities

    def call_genotype_probabilities(self, genotype_probabilities, threshold = 0.1):

        return ProbMath.call_genotype_probs(genotype_probabilities, threshold)


    def get_point_estimates(self, individual, haplotype_library, library_calling_threshold = 0.95, **kwargs) :
        called_haplotypes = haplotype_library.get_called_haplotypes(threshold = library_calling_threshold)
        mask = self.get_mask(called_haplotypes)     
        point_estimates = self.njit_get_point_estimates(individual.genotypes, called_haplotypes, self.error, mask)
        return point_estimates

    @staticmethod
    @jit(nopython=True, nogil=True)
    def njit_get_point_estimates(genotypes, haplotypes, error, mask):
        nHap, nLoci = haplotypes.shape
        point_estimates = np.full((nLoci, nHap), 1, dtype = np.float32)
        for i in range(nLoci):
            if genotypes[i] != 9 and mask[i]:
                for j in range(nHap):
                    sourceGeno = haplotypes[j, i]
                    if 2*sourceGeno == genotypes[i]:
                        point_estimates[i, j] = 1-error[i]
                    else:
                        point_estimates[i, j] = error[i]

        return point_estimates

    @staticmethod
    @jit(nopython=True, nogil=True)
    def transmission(cummulative_probabilities, previous_point_probability, recombination_rate, output):
        output[:] = cummulative_probabilities * previous_point_probability
        normalize(output)
        output[:] *= (1-recombination_rate)
        output[:] += recombination_rate


    def create_directional_smoothing(self) :
        transmission = self.transmission

        @jit(nopython=True, nogil=True)
        def directional_smoothing(point_estimate, recombination_rate, forward = False, backward = False):
            
            output = np.full(point_estimate.shape, 1, dtype = np.float32)
            n_loci = point_estimate.shape[0]

            if forward:
                start = 1
                stop = n_loci
                step = 1

            if backward:
                start = n_loci - 2
                stop = -1
                step = -1

            for i in range(start, stop, step):
                transmission(output[i-step,:], point_estimate[i - step,:], recombination_rate[i], output[i,:])

            return output

        return directional_smoothing

    def create_apply_smoothing(self):
        directional_smoothing = self.directional_smoothing

        @jit(nopython=True, nogil=True)
        def apply_smoothing(point_estimate, recombination_rate):
            """Calculate normalized state probabilities at each loci using the forward-backward algorithm"""

            est = ( point_estimate * 
                    directional_smoothing(point_estimate, recombination_rate, forward = True) *
                    directional_smoothing(point_estimate, recombination_rate, backward = True) )

            # Return normalized probabilities

            normalize_along_first_axis(est)
            return est

        return apply_smoothing

    def create_sampling_algorithm(self, selection_function):
        directional_smoothing = self.directional_smoothing
        transmission = self.transmission

        @jit(nopython=True, nogil=True)
        def sample_path(point_estimate, recombination_rate):
            """Calculate normalized state probabilities at each loci using the forward-backward algorithm"""

            # Right now using a matrix output; will improve later.
            n_loci = point_estimate.shape[0]

            output = np.full(point_estimate.shape, 0, dtype = np.float32)

            forward_and_point_estimate = point_estimate * directional_smoothing(point_estimate, recombination_rate, forward = True)

            # First index.
            selected_index = selection_function(forward_and_point_estimate[-1].ravel())
            output[- 1].ravel()[selected_index] = 1 # Set the output value at the selected_index to 1.

            # Always sample backward (for tradition mostly).
            locus_estimate = np.full(point_estimate[0].shape, 0, dtype = np.float32)
            matrix_ones = np.full(point_estimate[0].shape, 1, dtype = np.float32)

            start = n_loci - 2
            stop = -1
            step = -1

            for i in range(start, stop, step):
                # Pass along sampled value at the locus.
                transmission(output[i-step,:], matrix_ones, recombination_rate[i], locus_estimate)
                # Combine forward_estimate with backward_estimate
                locus_estimate *= forward_and_point_estimate[i,:]
                selected_index = selection_function(locus_estimate.ravel())
                output[i].ravel()[selected_index] = 1 # Set the output value at the selected_index to 1.

            # Return probabilities
            return output

        return sample_path


    def create_viterbi_algorithm(self):

        maximum_likelihood_step = self.maximum_likelihood_step
        @jit(nopython=True, nogil=True)
        def viterbi_path(point_estimate, recombination_rate):
            """Calculate normalized state probabilities at each loci using the forward-backward algorithm"""

            # Right now using a matrix output; will improve later.
            n_loci = point_estimate.shape[0]

            path_score = np.full(point_estimate.shape, 0, dtype = np.float32)
            previous_index = np.full(point_estimate.shape, 0, dtype = np.int64)

            output = np.full(point_estimate.shape, 0, dtype = np.float32)

            path_score[0] = point_estimate[0]
            start = 1; stop = n_loci; step = 1
            for i in range(start, stop, step):
                # Pass along sampled value at the locus.
                maximum_likelihood_step(path_score[i-step], recombination_rate[i], point_estimate[i], path_score[i], previous_index[i])

            # Traceback
            start_index = np.argmax(path_score[-1])
            output[n_loci-1].ravel()[start_index] = 1

            index = start_index
            start = n_loci-2; stop = -1; step = -1
            for i in range(start, stop, step):
                index = previous_index[i-step].ravel()[index]
                output[i].ravel()[index] = 1
            return output

        return viterbi_path

    @staticmethod
    @jit(nopython=True, nogil=True)
    def maximum_likelihood_step(previous_path_score, recombination_rate, point_estimate, output_path_score, output_index):

        best_index = np.argmax(previous_path_score)
        best_score = previous_path_score[best_index]

        n_hap = previous_path_score.shape[0]
        for i in range(n_hap):

            no_rec_score = (1-recombination_rate)*previous_path_score[i]
            rec_score = best_score*recombination_rate

            if no_rec_score > rec_score:
                # No recombination
                output_path_score[i] = no_rec_score*point_estimate[i]
                output_index[i] = i
            else:
                # Recombination
                output_path_score[i] = rec_score/n_hap*point_estimate[i]
                output_index[i] = best_index
        
        output_path_score /= np.sum(output_path_score)

    def calculate_genotype_probabilities(self, total_probs, haplotype_library, **kwargs):
        haplotype_dosages = haplotype_library.get_haplotypes()
        return self.njit_calculate_genotype_probabilities(total_probs, haplotype_dosages)


    @staticmethod        
    @jit(nopython=True, nogil=True)
    def njit_calculate_genotype_probabilities(total_probs, reference_haplotypes) :
        n_hap, n_loci = reference_haplotypes.shape
        geno_probs = np.full((2, n_loci), 0.0000001, dtype = np.float32) # Adding a very small value as a prior incase all of the values are missing.

        for i in range(n_loci):
            for j in range(n_hap):
                hap_value = reference_haplotypes[j, i]
                prob_value = total_probs[i,j]
                if hap_value != 9:
                    # Add in a sum of total_probs values. 
                    geno_probs[0, i] += prob_value * (1-hap_value)
                    geno_probs[1, i] += prob_value * hap_value

        geno_probs = geno_probs/np.sum(geno_probs, axis = 0)
        return geno_probs


class DiploidMarkovModel(HaploidMarkovModel) :
    def __init__(self, n_loci, error, recombination_rate = None):
        HaploidMarkovModel.__init__(self, n_loci, error, recombination_rate)


    def extract_reference_panels(self, haplotype_library = None, paternal_haplotype_library = None, maternal_haplotype_library = None) :
        if maternal_haplotype_library is not None and paternal_haplotype_library is not None:
            seperate_reference_panels = True
            return paternal_haplotype_library, maternal_haplotype_library, seperate_reference_panels

        else:
            seperate_reference_panels = False
            return haplotype_library, haplotype_library, seperate_reference_panels


    def get_point_estimates(self, individual, library_calling_threshold= 0.95, **kwargs):
        paternal_haplotype_library, maternal_haplotype_library, seperate_reference_panels = self.extract_reference_panels(**kwargs)

        paternal_called_haplotypes = paternal_haplotype_library.get_called_haplotypes(threshold = library_calling_threshold)
        maternal_called_haplotypes = maternal_haplotype_library.get_called_haplotypes(threshold = library_calling_threshold)

        mask = self.get_mask(paternal_called_haplotypes) & self.get_mask(maternal_called_haplotypes) 
        return self.njit_get_point_estimates(individual.genotypes, paternal_called_haplotypes, maternal_called_haplotypes, self.error, mask)


    @staticmethod
    @jit(nopython=True, nogil=True)
    def njit_get_point_estimates(indGeno, paternalHaplotypes, maternalHaplotypes, error, mask):
        nPat, nLoci = paternalHaplotypes.shape
        nMat, nLoci = maternalHaplotypes.shape

        point_estimates = np.full((nLoci, nPat, nMat), 1, dtype = np.float32)

        for i in range(nLoci):
            if indGeno[i] != 9 and mask[i]:
                for j in range(nPat):
                    for k in range(nMat):
                        sourceGeno = paternalHaplotypes[j, i] + maternalHaplotypes[k, i]
                        if sourceGeno == indGeno[i]:
                            point_estimates[i, j, k] = 1-error[i]
                        else:
                            point_estimates[i, j, k] = error[i]

        return point_estimates


    def calculate_genotype_probabilities(self, total_probs, haplotype_library = None, paternal_haplotype_library= None, maternal_haplotype_library= None, **kwargs):
        paternal_haplotype_library, maternal_haplotype_library, seperate_reference_panels = self.extract_reference_panels(haplotype_library, paternal_haplotype_library, maternal_haplotype_library)
        return self.njit_calculate_genotype_probabilities(total_probs, paternal_haplotype_library.get_haplotypes(), maternal_haplotype_library.get_haplotypes(), seperate_reference_panels)

    @staticmethod
    @jit(nopython=True, nogil=True)
    def njit_calculate_genotype_probabilities(total_probs, paternal_haplotypes, maternal_haplotypes, seperate_reference_panels) :
        n_pat, n_loci = paternal_haplotypes.shape
        n_mat, n_loci = maternal_haplotypes.shape
        geno_probs = np.full((4, n_loci), 0.00001, dtype = np.float32)

        for i in range(n_loci):
            for j in range(n_pat):
                for k in range(n_mat):
                    # diploid case where the markers are assumed independent.
                    if seperate_reference_panels or j != k: 
                        pat_value = paternal_haplotypes[j, i]
                        mat_value = maternal_haplotypes[k, i]
                        prob_value = total_probs[i,j,k]
                        if pat_value != 9 and mat_value != 9:
                            # Add in a sum of total_probs values. 
                            geno_probs[0, i] += prob_value * (1-pat_value)*(1-mat_value)   #aa
                            geno_probs[1, i] += prob_value * (1-pat_value)*mat_value     #aA
                            geno_probs[2, i] += prob_value * pat_value*(1-mat_value)     #Aa
                            geno_probs[3, i] += prob_value * pat_value*mat_value       #AA

                    # Haploid case where the markers are not independent
                    else:
                        hap_value = paternal_haplotypes[j, i]
                        prob_value = total_probs[i,j,k]
                        if hap_value != 9:
                            geno_probs[0, i] += prob_value * (1-hap_value)
                            geno_probs[1, i] += 0
                            geno_probs[2, i] += 0
                            geno_probs[3, i] += prob_value * hap_value

        geno_probs = geno_probs/np.sum(geno_probs, axis = 0)
        return geno_probs


    
    
    @staticmethod
    @jit(nopython=True, nogil=True)
    def transmission(cummulative_probabilities, previous_point_probability, recombination_rate, output):
        output[:] = cummulative_probabilities * previous_point_probability
        normalize(output)

        row_sums = np.sum(output, 0)
        col_sums = np.sum(output, 1)

        output[:] *= (1 - recombination_rate)**2 # No recombination on either chromosome.
        output[:] += np.expand_dims(row_sums, 0)/output.shape[0]*recombination_rate*(1-recombination_rate) # recombination on the maternal (second) chromosome)
        output[:] += np.expand_dims(col_sums, 1)/output.shape[1]*recombination_rate*(1-recombination_rate) # recombination on the paternal (first) chromosome)
        output[:] += recombination_rate**2/output.size # double recombination
    
    @staticmethod
    @jit(nopython=True, nogil=True)
    def maximum_likelihood_step(previous_path_score, recombination_rate, point_estimate, output_path_score, output_index):

        n_pat = previous_path_score.shape[0]
        n_mat = previous_path_score.shape[1]

        combined_max_index = np.argmax(previous_path_score)
        combined_max_score = previous_path_score.ravel()[combined_max_index] * recombination_rate**2/(n_mat*n_pat)

        paternal_max_index = np.full(n_pat, 0, dtype = np.int64)
        paternal_max_value = np.full(n_pat, 0, dtype = np.float32)

        maternal_max_index = np.full(n_mat, 0, dtype = np.int64)
        maternal_max_value = np.full(n_mat, 0, dtype = np.float32)

        # Recombination on the maternal side, paternal side is fixed
        for i in range(n_pat):
            index = np.argmax(previous_path_score[i,:])
            paternal_max_value[i] = previous_path_score[i, index] * (1-recombination_rate)*recombination_rate/n_mat
            paternal_max_index[i] = i*n_mat + index


        # Recombination on the paternal side, maternal side is fixed
        for j in range(n_mat):
            index = np.argmax(previous_path_score[:, j])
            maternal_max_value[j] = previous_path_score[index, j] * (1-recombination_rate)*recombination_rate/n_pat
            maternal_max_index[j] = index*n_mat + j

        for i in range(n_pat):
            for j in range(n_mat):

                best_score = (1-recombination_rate)**2*previous_path_score[i,j]
                best_index = i*n_mat + j

                # Paternal recombination 
                if paternal_max_value[i] > best_score:
                    best_score = paternal_max_value[i]
                    best_index = paternal_max_index[i]

                if maternal_max_value[j] > best_score:
                    best_score = maternal_max_value[j]
                    best_index = maternal_max_index[j]
                
                if combined_max_score > best_score:
                    best_score = combined_max_score
                    best_index = combined_max_index

                output_path_score[i,j] = best_score*point_estimate[i,j]
                output_index[i,j] = best_index
        
 
        output_path_score /= np.sum(output_path_score)

class JointMarkovModel(HaploidMarkovModel) :
    def __init__(self, n_loci, error, recombination_rate = None):
        HaploidMarkovModel.__init__(self, n_loci, error, recombination_rate)


    @staticmethod
    @jit(nopython=True, nogil=True)
    def njit_get_point_estimates(indGeno, haplotypes, error, mask):
        n_hap, n_loci = haplotypes.shape

        point_estimates = np.full((n_loci, n_hap, n_hap + 1), 1, dtype = np.float32)

        diploid_section = point_estimates[:,:,0:-1]
        haploid_section = point_estimates[:,:,-1]

        # Diploid point estimates/Emission probabilities

        for i in range(n_loci):
            if indGeno[i] != 9 and mask[i]:
                for j in range(n_hap):
                    for k in range(n_hap):
                        sourceGeno = haplotypes[j, i] + haplotypes[k, i]
                        if sourceGeno == indGeno[i]:
                            diploid_section[i, j, k] = 1-error[i]
                        else:
                            diploid_section[i, j, k] = error[i]

        # Diploid point estimates/Emission probabilities
        for i in range(n_loci):
            if indGeno[i] != 9 and mask[i]:
                for j in range(n_hap):
                        sourceGeno = 2*haplotypes[j, i]
                        if sourceGeno == indGeno[i]:
                            haploid_section[i, j] = 1-error[i]
                        else:
                            haploid_section[i, j] = error[i]

        return point_estimates


    @staticmethod
    @jit(nopython=True, nogil=True)
    def njit_calculate_genotype_probabilities(total_probs, reference_haplotypes) :
        n_hap, n_loci = reference_haplotypes.shape
        geno_probs = np.full((4, n_loci), 0.00001, dtype = np.float32)

        diploid_section = total_probs[:,:,0:-1]
        haploid_section = total_probs[:,:,-1]

        for i in range(n_loci):
            for j in range(n_hap):
                for k in range(n_hap):
                    # diploid case where the markers are assumed independent.
                    if j != k: 
                        pat_value = reference_haplotypes[j, i]
                        mat_value = reference_haplotypes[k, i]
                        prob_value = diploid_section[i,j,k]
                        if pat_value != 9 and mat_value != 9:
                            # Add in a sum of total_probs values. 
                            geno_probs[0, i] += prob_value * (1-pat_value)*(1-mat_value)   #aa
                            geno_probs[1, i] += prob_value * (1-pat_value)*mat_value     #aA
                            geno_probs[2, i] += prob_value * pat_value*(1-mat_value)     #Aa
                            geno_probs[3, i] += prob_value * pat_value*mat_value       #AA

                    # markers are not independent
                    else:
                        hap_value = reference_haplotypes[j, i]
                        prob_value = diploid_section[i,j,k]
                        if hap_value != 9:
                            geno_probs[0, i] += prob_value * (1-hap_value)
                            geno_probs[1, i] += 0
                            geno_probs[2, i] += 0
                            geno_probs[3, i] += prob_value * hap_value

        for i in range(n_loci):
            for j in range(n_hap):
                hap_value = reference_haplotypes[j, i]
                prob_value = haploid_section[i,j]
                if hap_value != 9:
                    geno_probs[0, i] += prob_value * (1-hap_value)
                    geno_probs[1, i] += 0
                    geno_probs[2, i] += 0
                    geno_probs[3, i] += prob_value * hap_value

        geno_probs = geno_probs/np.sum(geno_probs, axis = 0)
        return geno_probs


    
    
    @staticmethod
    @jit(nopython=True, nogil=True)
    def transmission(cummulative_probabilities, previous_point_probability, recombination_rate, output):
        
        output[:] = cummulative_probabilities * previous_point_probability
        normalize(output)

        diploid_section = output[:,0:-1]
        haploid_section = output[:,-1]

        diploid_weight = np.sum(diploid_section)
        haploid_weight = np.sum(haploid_section)

        row_sums = np.sum(diploid_section, 0)
        col_sums = np.sum(diploid_section, 1)

        diploid_section[:] *= (1 - recombination_rate)**2
        diploid_section[:] += np.expand_dims(row_sums, 0)/diploid_section.shape[0]*recombination_rate*(1-recombination_rate) # recombination on the maternal (second) chromosome)
        diploid_section[:] += np.expand_dims(col_sums, 1)/diploid_section.shape[1]*recombination_rate*(1-recombination_rate) # recombination on the paternal (first) chromosome)
        diploid_section[:] += diploid_weight*recombination_rate**2/diploid_section.size # double recombination

        haploid_section[:] *= (1 - recombination_rate)
        haploid_section[:] += haploid_weight*recombination_rate/haploid_section.size

        # loose the recombination to the haploid section and add the haploid recombination to diploid
        diploid_section[:] *= (1 - recombination_rate)
        diploid_section[:] += recombination_rate * haploid_weight/diploid_section.size

        # loose the recombination to the haploid section and add the haploid recombination to diploid
        haploid_section[:] *= (1 - recombination_rate)
        haploid_section[:] += recombination_rate * diploid_weight/haploid_section.size

    
    # @staticmethod
    # @jit(nopython=True, nogil=True)
    # def maximum_likelihood_step(previous_path_score, recombination_rate, point_estimate, output_path_score, output_index):

    #     n_pat = previous_path_score.shape[0]
    #     n_mat = previous_path_score.shape[1]

    #     combined_max_index = np.argmax(previous_path_score)
    #     combined_max_score = previous_path_score.ravel()[combined_max_index] * recombination_rate**2/(n_mat*n_pat)

    #     paternal_max_index = np.full(n_pat, 0, dtype = np.int64)
    #     paternal_max_value = np.full(n_pat, 0, dtype = np.float32)

    #     maternal_max_index = np.full(n_mat, 0, dtype = np.int64)
    #     maternal_max_value = np.full(n_mat, 0, dtype = np.float32)

    #     # Recombination on the maternal side, paternal side is fixed
    #     for i in range(n_pat):
    #         index = np.argmax(previous_path_score[i,:])
    #         paternal_max_value[i] = previous_path_score[i, index] * (1-recombination_rate)*recombination_rate/n_mat
    #         paternal_max_index[i] = i*n_mat + index


    #     # Recombination on the paternal side, maternal side is fixed
    #     for j in range(n_mat):
    #         index = np.argmax(previous_path_score[:, j])
    #         maternal_max_value[j] = previous_path_score[index, j] * (1-recombination_rate)*recombination_rate/n_pat
    #         maternal_max_index[j] = index*n_mat + j

    #     for i in range(n_pat):
    #         for j in range(n_mat):

    #             best_score = (1-recombination_rate)**2*previous_path_score[i,j]
    #             best_index = i*n_mat + j

    #             # Paternal recombination 
    #             if paternal_max_value[i] > best_score:
    #                 best_score = paternal_max_value[i]
    #                 best_index = paternal_max_index[i]

    #             if maternal_max_value[j] > best_score:
    #                 best_score = maternal_max_value[j]
    #                 best_index = maternal_max_index[j]
                
    #             if combined_max_score > best_score:
    #                 best_score = combined_max_score
    #                 best_index = combined_max_index

    #             output_path_score[i,j] = best_score*point_estimate[i,j]
    #             output_index[i,j] = best_index
        
 
    #     output_path_score /= np.sum(output_path_score)


@jit(nopython=True, nogil=True)
def normalize(mat):
    mat[:] /= np.sum(mat)    

@jit(nopython=True, nogil=True)
def normalize_along_first_axis(mat):
    for i in range(mat.shape[0]):
        normalize(mat[i,:])





import concurrent.futures
import itertools
import numpy as np
import math
from . import InputOutput

def convert_data_to_line(data_tuple, fmt) :
    idx, data = data_tuple
    return idx + ' ' + ' '.join(map(fmt, data)) + '\n'

def convert_data_to_line_plink(data_tuple, fmt):
    """Format data for PLINK plain text output"""
    idx, data = data_tuple
    return f"0 {idx} 0 0 0 0  {' '.join(data.astype(fmt))}\n"

def writeLines(fileName, data_list, fmt, converter=convert_data_to_line):
    # print(f"Writing results to: {fileName}")
    try:
        iothreads = InputOutput.args.iothreads
    except AttributeError as error:
        iothreads = 1

    with open(fileName, 'w+') as f:

        if iothreads > 1:
            with concurrent.futures.ProcessPoolExecutor(max_workers = iothreads) as executor: # The minus one is to account for the main thread.
                # Break up into small-ish chunks to reduce overall memory cost.
                # Hard code: splits into 1k individuals.
                # These get then split up into one chunk per thread.
                subsets = split_by(data_list, 1000)
                for subset in subsets:
                    for result in executor.map(converter, subset, itertools.repeat(fmt), chunksize=math.ceil(1000/iothreads)):
                        f.write(result)

        if iothreads <= 1:
            for data_tuple in data_list:
                result = converter(data_tuple, fmt)
                f.write(result)


def writeLinesPlinkPlainTxt(fileName, data_list):
    """Write lines in PLINK plain text format"""
    writeLines(fileName, data_list, fmt=str, converter=convert_data_to_line_plink)


def split_by(array, step):

    output = []
    i = 0
    while i*step < len(array):
        start = i*step
        stop = (i+1)*step
        output.append(array[start:stop])
        i += 1
    return output



def process_input_line(line, startsnp, stopsnp, dtype):
    parts = line.split(); 
    idx = parts[0]
    parts = parts[1:]

    if startsnp is not None :
        parts = parts[startsnp : stopsnp + 1] #Offset 1 for id and 2 for id + include stopsnp

    data=np.array([int(val) for val in parts], dtype = dtype)

    return (idx, data)


def process_input_line_plink(line, startsnp, stopsnp, dtype):
    """Proces a line from PLINK .ped file
    Fields:
    0      Family ID ('FID')
    1      Within-family ID ('IID'; cannot be '0')
    2      Within-family ID of father ('0' if father isn't in dataset)
    3      Within-family ID of mother ('0' if mother isn't in dataset)
    4      Sex code ('1' = male, '2' = female, '0' = unknown)
    5      Phenotype value ('1' = control, '2' = case, '-9'/'0'/non-numeric = missing data if case/control)
    6-end  Genotypes as pairs of alleles (A, C, G or T)
 
    At present, this extracts individual's identifier as the within-family ID
    """
    parts = line.split()
    idx = parts[1]         # Use within-family ID
    genotypes = parts[6:]
    if startsnp is not None:
        genotypes = genotypes[startsnp*2: stopsnp*2 + 2]  # Each locus is represented by two alleles
    data = np.array(genotypes, dtype=np.bytes_)
    return (idx, data)


def readLines(fileName, startsnp, stopsnp, dtype, processor=process_input_line):
    # print(f"Reading in file: {fileName}")

    try:
        iothreads = InputOutput.args.iothreads
    except AttributeError as error:
        iothreads = 1

    output = []
    with open(fileName) as f:

        if iothreads > 1:
            # This could be more efficient, but it's dwarfed by some of the other stuff in the program.
            # i.e. line is roughly the same size as the haplotypes (2 bytes per genotype value, i.e. (space)(value); and two values per haplotype.

            all_outputs = []
            lines = list(itertools.islice(f, 1000))
            while len(lines) > 0:
                with concurrent.futures.ProcessPoolExecutor(max_workers = iothreads) as executor:
                    chunk_output = executor.map(processor, lines, itertools.repeat(startsnp), itertools.repeat(stopsnp), itertools.repeat(dtype), chunksize=math.ceil(1000/iothreads))
                all_outputs.append(chunk_output)
                lines = list(itertools.islice(f, 1000))
            output = itertools.chain.from_iterable(all_outputs)

        if iothreads <= 1:
            for line in f:
                output.append(processor(line, startsnp = startsnp, stopsnp = stopsnp, dtype = dtype))

    return output


def readLinesPlinkPlainTxt(fileName, startsnp, stopsnp, dtype):
    """Read lines in PLINK plain text format"""
    return readLines(fileName, startsnp, stopsnp, dtype, processor=process_input_line_plink)
